@@ -3,41 +3,50 @@
 - raises HTTPErrors
 - creates REST API models
 """
-
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
-
-from collections import defaultdict
-from datetime import datetime, timedelta
-from functools import partial
+import asyncio
 import os
+from collections import defaultdict
+from datetime import datetime
+from datetime import timedelta
+from functools import partial
 
+from jupyter_client.multikernelmanager import AsyncMultiKernelManager
+from jupyter_client.multikernelmanager import MultiKernelManager
+from jupyter_client.session import Session
+from jupyter_core.paths import exists
 from tornado import web
 from tornado.concurrent import Future
-from tornado.ioloop import IOLoop, PeriodicCallback
-
-from jupyter_client.session import Session
-from jupyter_client.multikernelmanager import MultiKernelManager, AsyncMultiKernelManager
-from jupyter_core.paths import exists
-from traitlets import (Any, Bool, Dict, List, Unicode, TraitError, Integer,
-       Float, Instance, default, validate
-)
-
-from jupyter_server.utils import to_os_path, ensure_async
-from jupyter_server._tz import utcnow, isoformat
-from ipython_genutils.py3compat import getcwd
+from tornado.ioloop import IOLoop
+from tornado.ioloop import PeriodicCallback
+from traitlets import Any
+from traitlets import Bool
+from traitlets import default
+from traitlets import Dict
+from traitlets import Float
+from traitlets import Instance
+from traitlets import Integer
+from traitlets import List
+from traitlets import TraitError
+from traitlets import Unicode
+from traitlets import validate
 
+from jupyter_server._tz import isoformat
+from jupyter_server._tz import utcnow
 from jupyter_server.prometheus.metrics import KERNEL_CURRENTLY_RUNNING_TOTAL
+from jupyter_server.utils import ensure_async
+from jupyter_server.utils import to_os_path
 
 
 class MappingKernelManager(MultiKernelManager):
     """A KernelManager that handles
-     - File mapping
-     - HTTP error handling
-     - Kernel message filtering
+    - File mapping
+    - HTTP error handling
+    - Kernel message filtering
     """
 
-    @default('kernel_manager_class')
+    @default("kernel_manager_class")
     def _default_kernel_manager_class(self):
         return "jupyter_client.ioloop.IOLoopKernelManager"
 
@@ -47,21 +56,23 @@ class MappingKernelManager(MultiKernelManager):
 
     _kernel_connections = Dict()
 
+    _kernel_ports = Dict()
+
     _culler_callback = None
 
     _initialized_culler = False
 
-    @default('root_dir')
+    @default("root_dir")
     def _default_root_dir(self):
         try:
             return self.parent.root_dir
         except AttributeError:
-            return getcwd()
+            return os.getcwd()
 
-    @validate('root_dir')
+    @validate("root_dir")
     def _update_root_dir(self, proposal):
         """Do a bit of validation of the root dir."""
-        value = proposal['value']
+        value = proposal["value"]
         if not os.path.isabs(value):
             # If we receive a non-absolute path, make it absolute.
             value = os.path.abspath(value)
@@ -69,28 +80,38 @@ class MappingKernelManager(MultiKernelManager):
             raise TraitError("kernel root dir %r is not a directory" % value)
         return value
 
-    cull_idle_timeout = Integer(0, config=True,
+    cull_idle_timeout = Integer(
+        0,
+        config=True,
         help="""Timeout (in seconds) after which a kernel is considered idle and ready to be culled.
         Values of 0 or lower disable culling. Very short timeouts may result in kernels being culled
-        for users with poor network connections."""
+        for users with poor network connections.""",
     )
 
-    cull_interval_default = 300 # 5 minutes
-    cull_interval = Integer(cull_interval_default, config=True,
-        help="""The interval (in seconds) on which to check for idle kernels exceeding the cull timeout value."""
+    cull_interval_default = 300  # 5 minutes
+    cull_interval = Integer(
+        cull_interval_default,
+        config=True,
+        help="""The interval (in seconds) on which to check for idle kernels exceeding the cull timeout value.""",
     )
 
-    cull_connected = Bool(False, config=True,
+    cull_connected = Bool(
+        False,
+        config=True,
         help="""Whether to consider culling kernels which have one or more connections.
-        Only effective if cull_idle_timeout > 0."""
+        Only effective if cull_idle_timeout > 0.""",
     )
 
-    cull_busy = Bool(False, config=True,
+    cull_busy = Bool(
+        False,
+        config=True,
         help="""Whether to consider culling kernels which are busy.
-        Only effective if cull_idle_timeout > 0."""
+        Only effective if cull_idle_timeout > 0.""",
     )
 
-    buffer_offline_messages = Bool(True, config=True,
+    buffer_offline_messages = Bool(
+        True,
+        config=True,
         help="""Whether messages from kernels whose frontends have disconnected should be buffered in-memory.
 
         When True (default), messages are buffered and replayed on reconnect,
@@ -98,10 +119,12 @@ class MappingKernelManager(MultiKernelManager):
 
         Disable if long-running kernels will produce too much output while
         no frontends are connected.
-        """
+        """,
     )
 
-    kernel_info_timeout = Float(60, config=True,
+    kernel_info_timeout = Float(
+        60,
+        config=True,
         help="""Timeout for giving up on a kernel (in seconds).
 
         On starting and restarting kernels, we check whether the
@@ -110,43 +133,46 @@ class MappingKernelManager(MultiKernelManager):
         before being presumed dead.
         This affects the MappingKernelManager (which handles kernel restarts)
         and the ZMQChannelsHandler (which handles the startup).
-        """
+        """,
     )
 
     _kernel_buffers = Any()
-    @default('_kernel_buffers')
-    def _default_kernel_buffers(self):
-        return defaultdict(lambda: {'buffer': [], 'session_key': '', 'channels': {}})
 
-    last_kernel_activity = Instance(datetime,
-        help="The last activity on any kernel, including shutting down a kernel")
+    @default("_kernel_buffers")
+    def _default_kernel_buffers(self):
+        return defaultdict(lambda: {"buffer": [], "session_key": "", "channels": {}})
+
+    last_kernel_activity = Instance(
+        datetime,
+        help="The last activity on any kernel, including shutting down a kernel",
+    )
 
     def __init__(self, **kwargs):
         self.pinned_superclass = MultiKernelManager
         self.pinned_superclass.__init__(self, **kwargs)
         self.last_kernel_activity = utcnow()
 
-    allowed_message_types = List(trait=Unicode(), config=True,
+    allowed_message_types = List(
+        trait=Unicode(),
+        config=True,
         help="""White list of allowed kernel message types.
         When the list is empty, all message types are allowed.
-        """
+        """,
     )
 
-    allow_tracebacks = Bool(True, config=True, help=(
-        'Whether to send tracebacks to clients on exceptions.'
-    ))
+    allow_tracebacks = Bool(
+        True, config=True, help=("Whether to send tracebacks to clients on exceptions.")
+    )
 
     traceback_replacement_message = Unicode(
-        'An exception occurred at runtime, which is not shown due to security reasons.',
+        "An exception occurred at runtime, which is not shown due to security reasons.",
         config=True,
-        help=(
-            'Message to print when allow_tracebacks is False, and an exception occurs'
-        )
+        help=("Message to print when allow_tracebacks is False, and an exception occurs"),
     )
 
-    #-------------------------------------------------------------------------
+    # -------------------------------------------------------------------------
     # Methods for managing kernels and sessions
-    #-------------------------------------------------------------------------
+    # -------------------------------------------------------------------------
 
     def _handle_kernel_died(self, kernel_id):
         """notice that a kernel died"""
@@ -178,28 +204,29 @@ class MappingKernelManager(MultiKernelManager):
             The name identifying which kernel spec to launch. This is ignored if
             an existing kernel is returned, but it may be checked in the future.
         """
-        if kernel_id is None:
+        if kernel_id is None or kernel_id not in self:
             if path is not None:
-                kwargs['cwd'] = self.cwd_for_path(path)
+                kwargs["cwd"] = self.cwd_for_path(path)
+            if kernel_id is not None:
+                kwargs["kernel_id"] = kernel_id
             kernel_id = await ensure_async(self.pinned_superclass.start_kernel(self, **kwargs))
             self._kernel_connections[kernel_id] = 0
-            self.start_watching_activity(kernel_id)
+            fut = asyncio.ensure_future(self._finish_kernel_start(kernel_id))
+            if not getattr(self, "use_pending_kernels", None):
+                await fut
+            # add busy/activity markers:
+            kernel = self.get_kernel(kernel_id)
+            kernel.execution_state = "starting"
+            kernel.reason = ""
+            kernel.last_activity = utcnow()
             self.log.info("Kernel started: %s" % kernel_id)
             self.log.debug("Kernel args: %r" % kwargs)
-            # register callback for failed auto-restart
-            self.add_restart_callback(kernel_id,
-                lambda : self._handle_kernel_died(kernel_id),
-                'dead',
-            )
 
             # Increase the metric of number of kernels running
             # for the relevant kernel type by 1
-            KERNEL_CURRENTLY_RUNNING_TOTAL.labels(
-                type=self._kernels[kernel_id].kernel_name
-            ).inc()
+            KERNEL_CURRENTLY_RUNNING_TOTAL.labels(type=self._kernels[kernel_id].kernel_name).inc()
 
         else:
-            self._check_kernel_id(kernel_id)
             self.log.info("Using existing kernel: %s" % kernel_id)
 
         # Initialize culling if not already
@@ -208,6 +235,58 @@ class MappingKernelManager(MultiKernelManager):
 
         return kernel_id
 
+    async def _finish_kernel_start(self, kernel_id):
+        km = self.get_kernel(kernel_id)
+        if hasattr(km, "ready"):
+            try:
+                await km.ready
+            except Exception:
+                self.log.exception(km.ready.exception())
+                return
+
+        self._kernel_ports[kernel_id] = km.ports
+        self.start_watching_activity(kernel_id)
+        # register callback for failed auto-restart
+        self.add_restart_callback(
+            kernel_id,
+            lambda: self._handle_kernel_died(kernel_id),
+            "dead",
+        )
+
+    def ports_changed(self, kernel_id):
+        """Used by ZMQChannelsHandler to determine how to coordinate nudge and replays.
+
+        Ports are captured when starting a kernel (via MappingKernelManager).  Ports
+        are considered changed (following restarts) if the referenced KernelManager
+        is using a set of ports different from those captured at startup.  If changes
+        are detected, the captured set is updated and a value of True is returned.
+
+        NOTE: Use is exclusive to ZMQChannelsHandler because this object is a singleton
+        instance while ZMQChannelsHandler instances are per WebSocket connection that
+        can vary per kernel lifetime.
+        """
+        changed_ports = self._get_changed_ports(kernel_id)
+        if changed_ports:
+            # If changed, update captured ports and return True, else return False.
+            self.log.debug(f"Port change detected for kernel: {kernel_id}")
+            self._kernel_ports[kernel_id] = changed_ports
+            return True
+        return False
+
+    def _get_changed_ports(self, kernel_id):
+        """Internal method to test if a kernel's ports have changed and, if so, return their values.
+
+        This method does NOT update the captured ports for the kernel as that can only be done
+        by ZMQChannelsHandler, but instead returns the new list of ports if they are different
+        than those captured at startup.  This enables the ability to conditionally restart
+        activity monitoring immediately following a kernel's restart (if ports have changed).
+        """
+        # Get current ports and return comparison with ports captured at startup.
+        km = self.get_kernel(kernel_id)
+        if km.ports != self._kernel_ports[kernel_id]:
+            return km.ports
+        return None
+
     def start_buffering(self, kernel_id, session_key, channels):
         """Start buffering messages for a kernel
 
@@ -215,11 +294,11 @@ class MappingKernelManager(MultiKernelManager):
         ----------
         kernel_id : str
             The id of the kernel to stop buffering.
-        session_key: str
+        session_key : str
             The session_key, if any, that should get the buffer.
             If the session_key matches the current buffered session_key,
             the buffer will be returned.
-        channels: dict({'channel': ZMQStream})
+        channels : dict({'channel': ZMQStream})
             The zmq channels whose messages should be buffered.
         """
 
@@ -234,15 +313,15 @@ class MappingKernelManager(MultiKernelManager):
         self.stop_buffering(kernel_id)
         buffer_info = self._kernel_buffers[kernel_id]
         # record the session key because only one session can buffer
-        buffer_info['session_key'] = session_key
+        buffer_info["session_key"] = session_key
         # TODO: the buffer should likely be a memory bounded queue, we're starting with a list to keep it simple
-        buffer_info['buffer'] = []
-        buffer_info['channels'] = channels
+        buffer_info["buffer"] = []
+        buffer_info["channels"] = channels
 
         # forward any future messages to the internal buffer
         def buffer_msg(channel, msg_parts):
             self.log.debug("Buffering msg on %s:%s", kernel_id, channel)
-            buffer_info['buffer'].append((channel, msg_parts))
+            buffer_info["buffer"].append((channel, msg_parts))
 
         for channel, stream in channels.items():
             stream.on_recv(partial(buffer_msg, channel))
@@ -254,7 +333,7 @@ class MappingKernelManager(MultiKernelManager):
         ----------
         kernel_id : str
             The id of the kernel to stop buffering.
-        session_key: str, optional
+        session_key : str, optional
             The session_key, if any, that should get the buffer.
             If the session_key matches the current buffered session_key,
             the buffer will be returned.
@@ -264,7 +343,7 @@ class MappingKernelManager(MultiKernelManager):
             return
 
         buffer_info = self._kernel_buffers[kernel_id]
-        if buffer_info['session_key'] == session_key:
+        if buffer_info["session_key"] == session_key:
             # remove buffer
             self._kernel_buffers.pop(kernel_id)
             # only return buffer_info if it's a match
@@ -287,33 +366,37 @@ class MappingKernelManager(MultiKernelManager):
             return
         buffer_info = self._kernel_buffers.pop(kernel_id)
         # close buffering streams
-        for stream in buffer_info['channels'].values():
+        for stream in buffer_info["channels"].values():
             if not stream.closed():
                 stream.on_recv(None)
                 stream.close()
 
-        msg_buffer = buffer_info['buffer']
+        msg_buffer = buffer_info["buffer"]
         if msg_buffer:
-            self.log.info("Discarding %s buffered messages for %s",
-                len(msg_buffer), buffer_info['session_key'])
+            self.log.info(
+                "Discarding %s buffered messages for %s",
+                len(msg_buffer),
+                buffer_info["session_key"],
+            )
 
     def shutdown_kernel(self, kernel_id, now=False, restart=False):
         """Shutdown a kernel by kernel_id"""
         self._check_kernel_id(kernel_id)
-        kernel = self._kernels[kernel_id]
-        if kernel._activity_stream:
-            kernel._activity_stream.close()
-            kernel._activity_stream = None
+        self.stop_watching_activity(kernel_id)
         self.stop_buffering(kernel_id)
         self._kernel_connections.pop(kernel_id, None)
 
         # Decrease the metric of number of kernels
         # running for the relevant kernel type by 1
-        KERNEL_CURRENTLY_RUNNING_TOTAL.labels(
-            type=self._kernels[kernel_id].kernel_name
-        ).dec()
+        KERNEL_CURRENTLY_RUNNING_TOTAL.labels(type=self._kernels[kernel_id].kernel_name).dec()
 
-        return self.pinned_superclass.shutdown_kernel(self, kernel_id, now=now, restart=restart)
+        self.pinned_superclass.shutdown_kernel(self, kernel_id, now=now, restart=restart)
+        # Unlike its async sibling method in AsyncMappingKernelManager, removing the kernel_id
+        # from the connections dictionary isn't as problematic before the shutdown since the
+        # method is synchronous.  However, we'll keep the relative call orders the same from
+        # a maintenance perspective.
+        self._kernel_connections.pop(kernel_id, None)
+        self._kernel_ports.pop(kernel_id, None)
 
     async def restart_kernel(self, kernel_id, now=False):
         """Restart a kernel by kernel_id"""
@@ -329,7 +412,7 @@ class MappingKernelManager(MultiKernelManager):
             if not channel.closed():
                 channel.close()
             loop.remove_timeout(timeout)
-            kernel.remove_restart_callback(on_restart_failed, 'dead')
+            kernel.remove_restart_callback(on_restart_failed, "dead")
 
         def on_reply(msg):
             self.log.debug("Kernel info reply received: %s", kernel_id)
@@ -349,11 +432,15 @@ class MappingKernelManager(MultiKernelManager):
             if not future.done():
                 future.set_exception(RuntimeError("Restart failed"))
 
-        kernel.add_restart_callback(on_restart_failed, 'dead')
+        kernel.add_restart_callback(on_restart_failed, "dead")
         kernel.session.send(channel, "kernel_info_request")
         channel.on_recv(on_reply)
         loop = IOLoop.current()
         timeout = loop.add_timeout(loop.time() + self.kernel_info_timeout, on_timeout)
+        # Re-establish activity watching if ports have changed...
+        if self._get_changed_ports(kernel_id) is not None:
+            self.stop_watching_activity(kernel_id)
+            self.start_watching_activity(kernel_id)
         return future
 
     def notify_connect(self, kernel_id):
@@ -375,12 +462,14 @@ class MappingKernelManager(MultiKernelManager):
         kernel = self._kernels[kernel_id]
 
         model = {
-            "id":kernel_id,
+            "id": kernel_id,
             "name": kernel.kernel_name,
             "last_activity": isoformat(kernel.last_activity),
             "execution_state": kernel.execution_state,
-            "connections": self._kernel_connections[kernel_id],
+            "connections": self._kernel_connections.get(kernel_id, 0),
         }
+        if getattr(kernel, "reason", None):
+            model["reason"] = kernel.reason
         return model
 
     def list_kernels(self):
@@ -388,15 +477,18 @@ class MappingKernelManager(MultiKernelManager):
         kernels = []
         kernel_ids = self.pinned_superclass.list_kernel_ids(self)
         for kernel_id in kernel_ids:
-            model = self.kernel_model(kernel_id)
-            kernels.append(model)
+            try:
+                model = self.kernel_model(kernel_id)
+                kernels.append(model)
+            except (web.HTTPError, KeyError):
+                pass  # Probably due to a (now) non-existent kernel, continue building the list
         return kernels
 
     # override _check_kernel_id to raise 404 instead of KeyError
     def _check_kernel_id(self, kernel_id):
         """Check a that a kernel_id exists and raise 404 if not."""
         if kernel_id not in self:
-            raise web.HTTPError(404, u'Kernel does not exist: %s' % kernel_id)
+            raise web.HTTPError(404, "Kernel does not exist: %s" % kernel_id)
 
     # monitoring activity:
 
@@ -408,7 +500,8 @@ class MappingKernelManager(MultiKernelManager):
         """
         kernel = self._kernels[kernel_id]
         # add busy/activity markers:
-        kernel.execution_state = 'starting'
+        kernel.execution_state = "starting"
+        kernel.reason = ""
         kernel.last_activity = utcnow()
         kernel._activity_stream = kernel.connect_iopub()
         session = Session(
@@ -421,17 +514,30 @@ class MappingKernelManager(MultiKernelManager):
             self.last_kernel_activity = kernel.last_activity = utcnow()
 
             idents, fed_msg_list = session.feed_identities(msg_list)
-            msg = session.deserialize(fed_msg_list)
+            msg = session.deserialize(fed_msg_list, content=False)
 
-            msg_type = msg['header']['msg_type']
-            if msg_type == 'status':
-                kernel.execution_state = msg['content']['execution_state']
-                self.log.debug("activity on %s: %s (%s)", kernel_id, msg_type, kernel.execution_state)
+            msg_type = msg["header"]["msg_type"]
+            if msg_type == "status":
+                msg = session.deserialize(fed_msg_list)
+                kernel.execution_state = msg["content"]["execution_state"]
+                self.log.debug(
+                    "activity on %s: %s (%s)",
+                    kernel_id,
+                    msg_type,
+                    kernel.execution_state,
+                )
             else:
                 self.log.debug("activity on %s: %s", kernel_id, msg_type)
 
         kernel._activity_stream.on_recv(record_activity)
 
+    def stop_watching_activity(self, kernel_id):
+        """Stop watching IOPub messages on a kernel for activity."""
+        kernel = self._kernels[kernel_id]
+        if getattr(kernel, "_activity_stream", None):
+            kernel._activity_stream.close()
+            kernel._activity_stream = None
+
     def initialize_culler(self):
         """Start idle culler if 'cull_idle_timeout' is greater than zero.
 
@@ -440,14 +546,21 @@ class MappingKernelManager(MultiKernelManager):
         if not self._initialized_culler and self.cull_idle_timeout > 0:
             if self._culler_callback is None:
                 loop = IOLoop.current()
-                if self.cull_interval <= 0: #handle case where user set invalid value
-                    self.log.warning("Invalid value for 'cull_interval' detected (%s) - using default value (%s).",
-                        self.cull_interval, self.cull_interval_default)
+                if self.cull_interval <= 0:  # handle case where user set invalid value
+                    self.log.warning(
+                        "Invalid value for 'cull_interval' detected (%s) - using default value (%s).",
+                        self.cull_interval,
+                        self.cull_interval_default,
+                    )
                     self.cull_interval = self.cull_interval_default
                 self._culler_callback = PeriodicCallback(
-                    self.cull_kernels, 1000*self.cull_interval)
-                self.log.info("Culling kernels with idle durations > %s seconds at %s second intervals ...",
-                    self.cull_idle_timeout, self.cull_interval)
+                    self.cull_kernels, 1000 * self.cull_interval
+                )
+                self.log.info(
+                    "Culling kernels with idle durations > %s seconds at %s second intervals ...",
+                    self.cull_idle_timeout,
+                    self.cull_interval,
+                )
                 if self.cull_busy:
                     self.log.info("Culling kernels even if busy")
                 if self.cull_connected:
@@ -457,41 +570,69 @@ class MappingKernelManager(MultiKernelManager):
         self._initialized_culler = True
 
     async def cull_kernels(self):
-        self.log.debug("Polling every %s seconds for kernels idle > %s seconds...",
-            self.cull_interval, self.cull_idle_timeout)
+        self.log.debug(
+            "Polling every %s seconds for kernels idle > %s seconds...",
+            self.cull_interval,
+            self.cull_idle_timeout,
+        )
         """Create a separate list of kernels to avoid conflicting updates while iterating"""
         for kernel_id in list(self._kernels):
             try:
                 await self.cull_kernel_if_idle(kernel_id)
             except Exception as e:
-                self.log.exception("The following exception was encountered while checking the idle duration of kernel %s: %s",
-                    kernel_id, e)
+                self.log.exception(
+                    "The following exception was encountered while checking the idle duration of kernel %s: %s",
+                    kernel_id,
+                    e,
+                )
 
     async def cull_kernel_if_idle(self, kernel_id):
         kernel = self._kernels[kernel_id]
-        if hasattr(kernel, 'last_activity'):  # last_activity is monkey-patched, so ensure that has occurred
-            self.log.debug("kernel_id=%s, kernel_name=%s, last_activity=%s",
-                           kernel_id, kernel.kernel_name, kernel.last_activity)
+
+        if getattr(kernel, "execution_state") == "dead":
+            self.log.warning(
+                "Culling '%s' dead kernel '%s' (%s).",
+                kernel.execution_state,
+                kernel.kernel_name,
+                kernel_id,
+            )
+            await ensure_async(self.shutdown_kernel(kernel_id))
+            return
+
+        if hasattr(
+            kernel, "last_activity"
+        ):  # last_activity is monkey-patched, so ensure that has occurred
+            self.log.debug(
+                "kernel_id=%s, kernel_name=%s, last_activity=%s",
+                kernel_id,
+                kernel.kernel_name,
+                kernel.last_activity,
+            )
             dt_now = utcnow()
             dt_idle = dt_now - kernel.last_activity
             # Compute idle properties
             is_idle_time = dt_idle > timedelta(seconds=self.cull_idle_timeout)
-            is_idle_execute = self.cull_busy or (kernel.execution_state != 'busy')
+            is_idle_execute = self.cull_busy or (kernel.execution_state != "busy")
             connections = self._kernel_connections.get(kernel_id, 0)
             is_idle_connected = self.cull_connected or not connections
             # Cull the kernel if all three criteria are met
-            if (is_idle_time and is_idle_execute and is_idle_connected):
+            if is_idle_time and is_idle_execute and is_idle_connected:
                 idle_duration = int(dt_idle.total_seconds())
-                self.log.warning("Culling '%s' kernel '%s' (%s) with %d connections due to %s seconds of inactivity.",
-                                 kernel.execution_state, kernel.kernel_name, kernel_id, connections, idle_duration)
+                self.log.warning(
+                    "Culling '%s' kernel '%s' (%s) with %d connections due to %s seconds of inactivity.",
+                    kernel.execution_state,
+                    kernel.kernel_name,
+                    kernel_id,
+                    connections,
+                    idle_duration,
+                )
                 await ensure_async(self.shutdown_kernel(kernel_id))
 
 
 # AsyncMappingKernelManager inherits as much as possible from MappingKernelManager,
 # overriding only what is different.
 class AsyncMappingKernelManager(MappingKernelManager, AsyncMultiKernelManager):
-
-    @default('kernel_manager_class')
+    @default("kernel_manager_class")
     def _default_kernel_manager_class(self):
         return "jupyter_client.ioloop.AsyncIOLoopKernelManager"
 
@@ -503,19 +644,17 @@ class AsyncMappingKernelManager(MappingKernelManager, AsyncMultiKernelManager):
     async def shutdown_kernel(self, kernel_id, now=False, restart=False):
         """Shutdown a kernel by kernel_id"""
         self._check_kernel_id(kernel_id)
-        kernel = self._kernels[kernel_id]
-        if kernel._activity_stream:
-            kernel._activity_stream.close()
-            kernel._activity_stream = None
+        self.stop_watching_activity(kernel_id)
         self.stop_buffering(kernel_id)
 
         # Decrease the metric of number of kernels
         # running for the relevant kernel type by 1
-        KERNEL_CURRENTLY_RUNNING_TOTAL.labels(
-            type=self._kernels[kernel_id].kernel_name
-        ).dec()
+        KERNEL_CURRENTLY_RUNNING_TOTAL.labels(type=self._kernels[kernel_id].kernel_name).dec()
 
         # Finish shutting down the kernel before clearing state to avoid a race condition.
-        ret = await self.pinned_superclass.shutdown_kernel(self, kernel_id, now=now, restart=restart)
+        ret = await self.pinned_superclass.shutdown_kernel(
+            self, kernel_id, now=now, restart=restart
+        )
         self._kernel_connections.pop(kernel_id, None)
+        self._kernel_ports.pop(kernel_id, None)
         return ret
